@@ -91,7 +91,7 @@ def pe_section_flags(sec):
         x = 'x'
     return r + w + x
 
-# pylint: disable=too-many-arguments
+# pylint: disable=too-many-arguments,too-many-positional-arguments
 def pe_print_sigs(filename, pe, indent, extract, verbose, varlist = None):
     i  = f'{"":{indent}s}'
     ii = f'{"":{indent+3}s}'
@@ -151,6 +151,64 @@ def zboot_binary(pe, indent, verbose, varlist = None):
     except pefile.PEFormatError:
         print(f'# {i}      not a PE binary')
 
+def bzimage_binary(pe, indent):
+    i = f'{"":{indent}s}'
+    (jump, magic, vmin, vmaj, u1, u2, veroff) = struct.unpack_from('<H4sBBIHH',
+                                                                   pe.__data__[ 0x200 : ])
+    if magic != b'HdrS':
+        return
+    (u3, setup_secs) = struct.unpack_from('<BB', pe.__data__[ 0x1f0 : ])
+    (pl_off, pl_len) = struct.unpack_from('<II', pe.__data__[ 0x248 : ])
+
+    setup_size = setup_secs * 512 + 512
+    pl_off += setup_size
+    verstr = pedecode.getcstr(pe.__data__[ veroff + 0x200 : ]).decode()
+
+    (zm1, zm2, zm3, zm4) = struct.unpack_from('<BBBB', pe.__data__, pl_off)
+    if zm1 == 0x28 and zm2 == 0xb5 and zm3 == 0x2f and zm4 == 0xfd:
+        zalg = 'zstd'
+    elif zm1 == 0xfd and zm2 == 0x37 and zm3 == 0x7a and zm4 == 0x58:
+        zalg = 'xz'
+    elif zm1 == 0x1f and zm2 == 0x8b:
+        zalg = 'gzip'
+    else:
+        zalg = f'unknown ({zm1:x},{zm2:x},{zm3:x},{zm4:x})'
+
+    print(f'# {i}bzImage: bootver={vmaj}.{vmin} setup=0x{setup_size:x} '
+          f'payload=0x{pl_off:x}+0x{pl_len:x} compress={zalg}')
+    print(f'# {i}   {verstr}')
+
+def grub_type(mtype):
+    type2name = {
+        0 : 'elf',
+        1 : 'memdisk',
+        2 : 'config',
+        3 : 'prefix',
+        4 : 'pubkey',
+        5 : 'dtb',
+        6 : 'disable-shim-lock',
+        7 : 'disable-cli',
+    }
+    return type2name.get(mtype, f'{mtype}')
+
+def grub_mods(sec, indent):
+    i = f'{"":{indent}s}'
+    blob = sec.get_data()
+    (magic, offset32) = struct.unpack_from('<LL', blob)
+    if magic != 0x676d696d:
+        return
+    if offset32:
+        (magic, offset, size) = struct.unpack_from('<LLL', blob)
+    else:
+        (magic, padding, offset, size) = struct.unpack_from('<LLQQ', blob)
+    print(f'# {i}grub modules: offset=0x{offset:x}, size=0x{size:x}')
+    pos = offset
+    while pos + 8 < size:
+        (mtype, msize) = struct.unpack_from('<LL', blob, pos)
+        typename = grub_type(mtype)
+        print(f'# {i}   module: type={typename}, offset=0x{pos:x}, size=0x{msize:x}')
+        pos += msize
+
 # pylint: disable=too-many-branches
 def pe_print_section(pe, sec, indent, verbose, varlist = None):
     i  = f'{"":{indent}s}'
@@ -192,9 +250,13 @@ def pe_print_section(pe, sec, indent, verbose, varlist = None):
             for nsec in npe.sections:
                 pe_print_section(npe, nsec, indent + 6, verbose, varlist)
             zboot_binary(npe, indent + 6, verbose, varlist)
+            bzimage_binary(pe, 3)
             pe_print_sigs(None, npe, indent + 6, False, verbose, varlist)
         except pefile.PEFormatError:
             print(f'# {ii}   not a PE binary')
+    if secname == b'mods':
+        if verbose:
+            grub_mods(sec, indent+3)
 
 def efi_binary(filename, extract = False, verbose = False, varlist = None):
     print(f'# file: {filename}')
@@ -204,6 +266,7 @@ def efi_binary(filename, extract = False, verbose = False, varlist = None):
         for sec in pe.sections:
             pe_print_section(pe, sec, 3, verbose)
         zboot_binary(pe, 3, verbose)
+        bzimage_binary(pe, 3)
         pe_print_sigs(filename, pe, 3, extract, verbose, varlist)
     except pefile.PEFormatError:
         print('#    not a PE binary')
